/* Copyright (C) 2001 2002 2003 Chris Vine

This program is distributed under the General Public Licence, version 2.
For particulars of this and relevant disclaimers see the file
COPYING distributed with the source files.

*/


#include <vector>

#include <cstdlib>
#include <fstream>
#include <iomanip>
#include <list>
#include <cstring>

#include <unistd.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <dirent.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#ifdef HAVE_PTHREAD_SIGMASK 
#include <pthread.h>
#endif

#include <gdkmm/pixbuf.h>
#include <gdk/gdkkeysyms.h> // the key codes are here
#include <gtkmm/main.h>
#include <gtkmm/image.h>
#include <gtkmm/stock.h>
#include <gtkmm/tooltips.h>
#include <glibmm/convert.h>
#include <glibmm/timer.h>
#include <glibmm/thread.h>

#include "fax_list.h"
#include "dialogs.h"
#include "fax_list_icons.h"
#include "window_icon.h"
#include "pipes.h"

#ifdef HAVE_STRINGSTREAM
#include <sstream>
#else
#include <strstream>
#endif

#ifdef ENABLE_NLS
#include <libintl.h>
#endif

int FaxListDialog::is_fax_received_list = 0;
int FaxListDialog::is_fax_sent_list = 0;


FaxListDialog::FaxListDialog(Mode mode_, const int standard_size_):
                              Gtk::Window(Gtk::WINDOW_TOPLEVEL),
			      mode(mode_), standard_size(standard_size_),
			      fax_basename_sem(1),
			      fax_list_box(false, 0), table(2, 1, false),
                              close_button(Gtk::Stock::CLOSE),
			      button_box(Gtk::BUTTONBOX_SPREAD) {

  // notify the existence of this object
  if (mode == received) is_fax_received_list++;
  else is_fax_sent_list++;

  fax_list_scroll_window.set_policy(Gtk::POLICY_ALWAYS, Gtk::POLICY_ALWAYS);
  fax_list_scroll_window.add(tree_view);
  fax_list_scroll_window.set_shadow_type(Gtk::SHADOW_IN);

  // create the tree model:
  list_store_r = Gtk::ListStore::create(model_columns);
  // connect it to the tree view
  tree_view.set_model(list_store_r);
  // add the columns of the tree model to tree view
  tree_view.append_column(gettext("Fax"), model_columns.fax_name);
  tree_view.append_column(gettext("Description"), model_columns.fax_description);
  // single line selection
  tree_view.get_selection()->set_mode(Gtk::SELECTION_SINGLE);
  // populate the fax list
  get_fax_list_rows();

  // set up the tool bar
  tool_bar.set_orientation(Gtk::ORIENTATION_HORIZONTAL);
  tool_bar.set_toolbar_style(Gtk::TOOLBAR_ICONS);

  // first make the image widgets
  Gtk::Image* print_image_p = manage(new Gtk::Image(Gdk::Pixbuf::create_from_xpm_data(print_xpm)));
  Gtk::Image* view_image_p = manage(new Gtk::Image(Gdk::Pixbuf::create_from_xpm_data(view_xpm)));
  Gtk::Image* describe_image_p = manage(new Gtk::Image(Gdk::Pixbuf::create_from_xpm_data(describe_xpm)));
  Gtk::Image* delete_image_p = manage(new Gtk::Image(Gdk::Pixbuf::create_from_xpm_data(delete_xpm)));
  Gtk::Image* refresh_image_p = manage(new Gtk::Image(Gdk::Pixbuf::create_from_xpm_data(refresh_xpm)));

  tool_bar.set_tooltips(true);

#if GTKMM_VERSION > 20

  tool_bar.set_toolbar_style(Gtk::TOOLBAR_ICONS);

  Gtk::Tooltips* tooltips_p = tool_bar.get_tooltips_object();
  tooltips_p->enable();

  print_button_p = manage(new Gtk::ToolButton(*print_image_p));
  tool_bar.append(*print_button_p, sigc::mem_fun(*this, &FaxListDialog::print_fax_prompt));
  print_button_p->set_sensitive(false);
  print_button_p->set_tooltip(*tooltips_p, gettext("Print selected fax"));

  view_button_p = manage(new Gtk::ToolButton(*view_image_p));
  tool_bar.append(*view_button_p, sigc::mem_fun(*this, &FaxListDialog::view_fax));
  view_button_p->set_sensitive(false);
  view_button_p->set_tooltip(*tooltips_p, gettext("View selected fax"));

  describe_button_p = manage(new Gtk::ToolButton(*describe_image_p));
  tool_bar.append(*describe_button_p, sigc::mem_fun(*this, &FaxListDialog::describe_fax_prompt));
  describe_button_p->set_sensitive(false);
  describe_button_p->set_tooltip(*tooltips_p, gettext("Add/amend description for selected fax"));

  delete_button_p = manage(new Gtk::ToolButton(*delete_image_p));
  tool_bar.append(*delete_button_p, sigc::mem_fun(*this, &FaxListDialog::delete_fax_prompt));
  delete_button_p->set_sensitive(false);
  delete_button_p->set_tooltip(*tooltips_p, gettext("Delete selected fax"));

  Gtk::ToolButton* refresh_button_p = manage(new Gtk::ToolButton(*refresh_image_p));
  tool_bar.append(*refresh_button_p, sigc::mem_fun(*this, &FaxListDialog::refresh_slot));
  refresh_button_p->set_tooltip(*tooltips_p, gettext("Refresh fax list"));

#else

  Gtk::Button* refresh_button_p;
  {
    using namespace Gtk::Toolbar_Helpers;
    
    ToolList& tool_list = tool_bar.tools();
    tool_list.push_back(ButtonElem(*print_image_p, SigC::slot(*this, &FaxListDialog::print_fax_prompt),
				   gettext("Print selected fax")));
    print_button_p = static_cast<Gtk::Button*>(tool_list.back().get_widget());
    print_button_p->set_relief(Gtk::RELIEF_NONE);
    print_button_p->set_sensitive(false);

    tool_list.push_back(ButtonElem(*view_image_p, SigC::slot(*this, &FaxListDialog::view_fax),
				   gettext("View selected fax")));
    view_button_p = static_cast<Gtk::Button*>(tool_list.back().get_widget());
    view_button_p->set_relief(Gtk::RELIEF_NONE);
    view_button_p->set_sensitive(false);

    tool_list.push_back(ButtonElem(*describe_image_p, SigC::slot(*this, &FaxListDialog::describe_fax_prompt),
				   gettext("Add/amend description for selected fax")));
    describe_button_p = static_cast<Gtk::Button*>(tool_list.back().get_widget());
    describe_button_p->set_relief(Gtk::RELIEF_NONE);
    describe_button_p->set_sensitive(false);

    tool_list.push_back(ButtonElem(*delete_image_p, SigC::slot(*this, &FaxListDialog::delete_fax_prompt),
				   gettext("Delete selected fax")));
    delete_button_p = static_cast<Gtk::Button*>(tool_list.back().get_widget());
    delete_button_p->set_relief(Gtk::RELIEF_NONE);
    delete_button_p->set_sensitive(false);

    tool_list.push_back(ButtonElem(*refresh_image_p, SigC::slot(*this, &FaxListDialog::refresh_slot),
				   gettext("Refresh fax list")));
    refresh_button_p = static_cast<Gtk::Button*>(tool_list.back().get_widget());
  }

#endif

  button_box.add(close_button);

  table.attach(fax_list_scroll_window, 0, 1, 0, 1, Gtk::EXPAND | Gtk::FILL,
         Gtk::EXPAND | Gtk::FILL, standard_size/3, standard_size/3);
  table.attach(button_box, 0, 1, 1, 2, Gtk::EXPAND | Gtk::FILL,
	 Gtk::SHRINK, standard_size/3, standard_size/3);

  fax_list_box.pack_start(tool_bar, false, false);
  fax_list_box.pack_start(table, true, true);

  // now connect up the signal which indicates a selection has been made
  tree_view.get_selection()->signal_changed().connect(sigc::mem_fun(*this, &FaxListDialog::set_buttons_slot));

  // connect up close_button
  close_button.signal_clicked().connect(sigc::mem_fun(*this, &FaxListDialog::close_slot));
  close_button.set_flags(Gtk::CAN_DEFAULT);
  close_button.grab_focus();

  table.set_border_width(standard_size/3);

  if (mode == received) set_title(gettext("efax-gtk: Received fax list"));
  else set_title(gettext("efax-gtk: Sent fax list"));
  //set_position(Gtk::WIN_POS_CENTER);
  add(fax_list_box);

  set_default_size(standard_size * 15, standard_size * 12);

  set_icon(Gdk::Pixbuf::create_from_xpm_data(window_icon_xpm));

  show_all();

#if GTKMM_VERSION == 20
  refresh_button_p->set_relief(Gtk::RELIEF_NORMAL);
#endif
  refresh_button_p->set_sensitive(true);
}

FaxListDialog::~FaxListDialog(void) {
  // notify the destruction of this object
  if (mode == received) is_fax_received_list--;
  else is_fax_sent_list--;
}

void FaxListDialog::get_fax_list_rows(void) {

  std::string dir(prog_config.homedir);
  if (mode == received) dir += "/faxin";
  else dir += "/faxsent";

  chdir(dir.c_str());

  DIR* dir_p;
  if ((dir_p = opendir(dir.c_str())) == 0) {
    std::string msg("Can't open directory ");
    msg += dir;
    msg += '\n';
    write_error(msg.c_str());
  }
  else {

    struct dirent* direntry;
    struct stat statinfo;
    std::list<std::string> dir_list;

    // first populate dir_list with the directory names in $HOME/faxin
    while ((direntry = readdir(dir_p)) != 0) {
      stat(direntry->d_name, &statinfo);
      if (S_ISDIR(statinfo.st_mode)
	  && std::strcmp(direntry->d_name, "oldfax")
	  && std::strcmp(direntry->d_name, ".")
	  && std::strcmp(direntry->d_name, "..")
	  && (mode == sent
	      || std::strcmp(direntry->d_name, prog_config.receive_dirname))) {
	dir_list.push_back(direntry->d_name);
      }
    }
    closedir(dir_p);

    // now insert the directory names in fax_list, with description (if any)
    // first clear the list store -- we need to clear it here before
    // testing for an empty dir_list, or on deleting the last fax in the list,
    // the clear won't take out the former last entry

    list_store_r->clear();

    if (!dir_list.empty()) {
      // first sort them
      dir_list.sort();
      
      // populate the list store, and get the fax description (if any)
      std::string filename;
      std::string line;
      std::ifstream file;
      std::list<std::string>::const_iterator iter;

      for (iter = dir_list.begin(); iter!= dir_list.end(); ++iter) {
      
	// get a list store row to insert the fax name
	Gtk::TreeModel::Row row = *(list_store_r->append());
	row[model_columns.fax_name] = *iter;

	// now see if there is a description
	filename = dir + '/';
	filename += *iter;
	filename += "/Description";

#ifdef HAVE_IOS_NOCREATE
	file.open(filename.c_str(), std::ios::in | std::ios::nocreate);
#else
	// we must have Std C++ so we probably don't need a ios::nocreate
	// flag on a read open to ensure uniqueness
	file.open(filename.c_str(), std::ios::in);
#endif

	if (file) {
	  while (std::getline(file, line) && line.empty());
        }
	if (!line.empty()) {
	  try {
	    row[model_columns.fax_description] = Glib::locale_to_utf8(line);
	  }
	  catch (Glib::ConvertError&) {
	    write_error("UTF-8 conversion error in FaxListDialog::get_fax_list_rows()\n");
	  }
	}

	filename = "";
	line = "";
	file.clear();
	file.close();
      }
    }
  }
  // reset current directory
  std::string temp(prog_config.homedir + "/faxin");
  chdir(temp.c_str());
}

void FaxListDialog::close_slot(void) {
  hide_all();
  delete this; // this is completely safe as the dialog is self-owning and modeless
}


bool FaxListDialog::on_delete_event(GdkEventAny*) {
  close_slot();
  return true; // returning true prevents destroy sig being emitted
}

void FaxListDialog::set_buttons_slot(void) {

  // see if anything is selected
  if (tree_view.get_selection()->get_selected()) {
    print_button_p->set_sensitive(true);
    view_button_p->set_sensitive(true);
    describe_button_p->set_sensitive(true);
    delete_button_p->set_sensitive(true);

#if GTKMM_VERSION == 20
    print_button_p->set_relief(Gtk::RELIEF_NORMAL);
    view_button_p->set_relief(Gtk::RELIEF_NORMAL);
    describe_button_p->set_relief(Gtk::RELIEF_NORMAL);
    delete_button_p->set_relief(Gtk::RELIEF_NORMAL);
#endif
  }
    
  else {
    print_button_p->set_sensitive(false);
    view_button_p->set_sensitive(false);
    describe_button_p->set_sensitive(false);
    delete_button_p->set_sensitive(false);

#if GTKMM_VERSION == 20
    print_button_p->set_relief(Gtk::RELIEF_NONE);
    view_button_p->set_relief(Gtk::RELIEF_NONE);
    describe_button_p->set_relief(Gtk::RELIEF_NONE);
    delete_button_p->set_relief(Gtk::RELIEF_NONE);
#endif
  }
}

void FaxListDialog::describe_fax_prompt(void) {


  Gtk::TreeModel::iterator row_iter = tree_view.get_selection()->get_selected();
  if (row_iter) {
    Glib::ustring description;
    description = (*row_iter)[model_columns.fax_description];
    
    DescriptionDialog* dialog_p = new DescriptionDialog(standard_size, description, *this);
    dialog_p->accepted.connect(sigc::mem_fun(*this, &FaxListDialog::describe_fax));
    // there is no memory leak -- the memory will be deleted when DescriptionDialog closes
  }
}

void FaxListDialog::describe_fax(const Glib::ustring& description) {

  Gtk::TreeModel::iterator row_iter = tree_view.get_selection()->get_selected();
  if (row_iter) {
    (*row_iter)[model_columns.fax_description] = description;

    std::string filename(prog_config.homedir);
    if (mode == received) filename += "/faxin/";
    else filename += "/faxsent/";

    filename += (*row_iter)[model_columns.fax_name];
    filename += "/Description";
    std::ofstream file(filename.c_str(), std::ios::out);
    // this try()/catch() block is ultra-cautious - something has gone seriously
    // wrong with the program if the UTF-8 conversion fails since the source is
    // a Gtk::Entry object
    try {
      if (file) file << Glib::locale_from_utf8(description);
    
      else {
	std::string msg("Can't open file ");
	msg += filename;
	msg += '\n';
	write_error(msg.c_str());
      }
    }
    catch (Glib::ConvertError&) {
      write_error("UTF-8 conversion error in FaxListDialog::describe_fax()\n");
    }
  }
}

void FaxListDialog::print_fax_prompt(void) {

  if (prog_config.print_popup) {
    Gtk::TreeModel::iterator row_iter = tree_view.get_selection()->get_selected();
    if (row_iter && !prog_config.print_cmd.empty()) {
      std::string faxnumber((*row_iter)[model_columns.fax_name]);
      Glib::ustring msg(gettext("Print fax "));
      msg += faxnumber + gettext("?");
      
      PromptDialog* dialog_p = new PromptDialog(msg, gettext("Print fax"), standard_size, *this);
      dialog_p->accepted.connect(sigc::mem_fun(*this, &FaxListDialog::print_fax));
      // there is no memory leak -- the memory will be deleted when PromptDialog closes
    }
  }
  else print_fax();
}

std::pair<const char*, char* const*> FaxListDialog::get_print_from_stdin_parms(void) {

  // lock the Prog_config object to stop it being modified in the intial (GUI) thread
  // while we are accessing it here
  Glib::Mutex::Lock lock(*prog_config.mutex_p);
  std::vector<std::string> print_parms;
  std::string print_cmd;
  std::string print_name;
  std::string::size_type end_pos;
  try {
    print_cmd = Glib::filename_from_utf8(prog_config.print_cmd);
  }
  catch (Glib::ConvertError&) {
    write_error("UTF-8 conversion error in FaxListDialog::print_from_stdin()\n");
    return std::pair<const char*, char* const*>(0,0);
  }
  
  if ((end_pos = print_cmd.find_first_of(' ')) != std::string::npos) { // we have parms
    print_name.assign(print_cmd, 0, end_pos);
    print_parms.push_back(print_name);
    // find start of next parm
    std::string::size_type start_pos = print_cmd.find_first_not_of(' ', end_pos);
    while (start_pos != std::string::npos) {
      end_pos = print_cmd.find_first_of(' ', start_pos);
      if (end_pos != std::string::npos) {
	print_parms.push_back(print_cmd.substr(start_pos, end_pos - start_pos));
	start_pos = print_cmd.find_first_not_of(' ', end_pos); // prepare for next interation
      }
      else {
	print_parms.push_back(print_cmd.substr(start_pos, 
					       print_cmd.size() - start_pos));
	start_pos = end_pos;
      }
    }
  }

  else { // just a print command without parameters to be passed
    print_name = print_cmd;
    print_parms.push_back(print_name);
  }

  char** exec_parms = new char*[print_parms.size() + 1];

  char** temp_pp = exec_parms;
  std::vector<std::string>::const_iterator iter;
  for (iter = print_parms.begin(); iter != print_parms.end(); ++iter, ++temp_pp) {
    *temp_pp = new char[iter->size() + 1];
    std::strcpy(*temp_pp, iter->c_str());
  }
  
  *temp_pp = 0;

  char* prog_name = new char[print_name.size() + 1];
  std::strcpy(prog_name, print_name.c_str());

  return std::pair<const char*, char* const*>(prog_name, exec_parms);
}

std::pair<const char*, char* const*> FaxListDialog::get_fax_to_ps_parms
                                       (const std::string& basename, bool allow_shrink) {

  // lock the Prog_config object to stop it being modified in the intial (GUI) thread
  // while we are accessing it here
  Glib::Mutex::Lock lock(*prog_config.mutex_p);

  // set up the parms for efix
  std::vector<std::string> efix_parms;
  std::string temp;
    
  efix_parms.push_back("efix-0.9a");
  // shut up efix (comment out next line and uncomment following one if errors to be reported)
  efix_parms.push_back("-v");
  //efix_parms.push_back("-ve");
  efix_parms.push_back("-r300");
  efix_parms.push_back("-ops");
  temp = "-p";
  temp += prog_config.page_dim;
  efix_parms.push_back(temp);

#ifdef HAVE_STRINGSTREAM
  if (allow_shrink && prog_config.print_shrink.compare("100")) { // if print_shrink is not 100
    temp = "-s0.";
    temp += prog_config.print_shrink;
    efix_parms.push_back(temp);

    std::ostringstream strm;
    if (!prog_config.page_size.compare("a4")) {

      strm << "-d";
      float val = 210 * (100 - std::atoi(prog_config.print_shrink.c_str()))/200.0;
      strm << static_cast<int>(val + 0.5);
      val = 297 * (100 - std::atoi(prog_config.print_shrink.c_str()))/200.0;
      strm << ',' << static_cast<int>(val + 0.5) << "mm";
      
      efix_parms.push_back(strm.str());
    }
    
    else if (!prog_config.page_size.compare("letter")) {

      strm << "-d";
      float val = 216 * (100 - std::atoi(prog_config.print_shrink.c_str()))/200.0;
      strm << static_cast<int>(val + 0.5);
      val = 279 * (100 - std::atoi(prog_config.print_shrink.c_str()))/200.0;
      strm << ',' << static_cast<int>(val + 0.5) << "mm";
      
      efix_parms.push_back(strm.str());
    }

    else if (!prog_config.page_size.compare("legal")) {

      strm << "-d";
      float val = 216 * (100 - std::atoi(prog_config.print_shrink.c_str()))/200.0;
      strm << static_cast<int>(val + 0.5);
      val = 356 * (100 - std::atoi(prog_config.print_shrink.c_str()))/200.0;
      strm << ',' << static_cast<int>(val + 0.5) << "mm";
      
      efix_parms.push_back(strm.str());
    }
  }

  int partnumber = 1;
  std::ostringstream strm;
  strm << basename.c_str() << std::setfill('0')
       << std::setw(3) << partnumber;
  int result = access(strm.str().c_str(), R_OK);
    
  while (!result) {
    efix_parms.push_back(strm.str());
	
    partnumber++;
    strm.str("");
    strm << basename.c_str() << std::setfill('0')
	 << std::setw(3) << partnumber;
    result = access(strm.str().c_str(), R_OK);
  }

#else
  if (allow_shrink && prog_config.print_shrink.compare("100")) { // if print_shrink is not 100
    temp = "-s0.";
    temp += prog_config.print_shrink;
    efix_parms.push_back(temp);

    std::ostrstream strm;
    if (!prog_config.page_size.compare("a4")) {

      strm << "-d";
      float val = 210 * (100 - std::atoi(prog_config.print_shrink.c_str()))/200.0;
      strm << static_cast<int>(val + 0.5);
      val = 297 * (100 - std::atoi(prog_config.print_shrink.c_str()))/200.0;
      strm << ',' << static_cast<int>(val + 0.5) << "mm" << std::ends;
      
      const char* dim = strm.str();
      efix_parms.push_back(dim);
      delete[] dim;
    }
    
    else if (!prog_config.page_size.compare("letter")) {

      strm << "-d";
      float val = 216 * (100 - std::atoi(prog_config.print_shrink.c_str()))/200.0;
      strm << static_cast<int>(val + 0.5);
      val = 279 * (100 - std::atoi(prog_config.print_shrink.c_str()))/200.0;
      strm << ',' << static_cast<int>(val + 0.5) << "mm" << std::ends;
      
      const char* dim = strm.str();
      efix_parms.push_back(dim);
      delete[] dim;
    }

    else if (!prog_config.page_size.compare("legal")) {

      strm << "-d";
      float val = 216 * (100 - std::atoi(prog_config.print_shrink.c_str()))/200.0;
      strm << static_cast<int>(val + 0.5);
      val = 356 * (100 - std::atoi(prog_config.print_shrink.c_str()))/200.0;
      strm << ',' << static_cast<int>(val + 0.5) << "mm" << std::ends;
      
      const char* dim = strm.str();
      efix_parms.push_back(dim);
      delete[] dim;
    }
  }

  int partnumber = 1;
  std::ostrstream strm;
  strm << basename.c_str() << std::setfill('0')
       << std::setw(3) << partnumber << std::ends;
  const char* file_name = strm.str();
  int result = access(file_name, R_OK);
    
  while (!result) {
    efix_parms.push_back(file_name);
    delete[] file_name;
	
    partnumber++;
    std::ostrstream strm;
    strm << basename.c_str() << std::setfill('0')
	 << std::setw(3) << partnumber << std::ends;
    file_name = strm.str();
    result = access(file_name, R_OK);
  }
  delete[] file_name;
#endif

  char** exec_parms = new char*[efix_parms.size() + 1];

  char** temp_pp = exec_parms;
  std::vector<std::string>::const_iterator iter;
  for (iter = efix_parms.begin(); iter != efix_parms.end(); ++iter, ++temp_pp) {
    *temp_pp = new char[iter->size() + 1];
    std::strcpy(*temp_pp, iter->c_str());
  }
  
  *temp_pp = 0;

  char* prog_name = new char[std::strlen("efix-0.9a") + 1];
  std::strcpy(prog_name, "efix-0.9a");

  return std::pair<const char*, char* const*>(prog_name, exec_parms);
}

void FaxListDialog::write_from_stdin(int dest_fd) {

  ssize_t read_result;
  ssize_t write_result;
  char buffer[PIPE_BUF];

  while ((read_result = read(0, buffer, PIPE_BUF)) != 0) {
    if (read_result == -1) {
      if (errno == EINTR) continue;
      else break;
    }
    do {
      write_result = write(dest_fd, buffer, read_result);
    } while (write_result == -1 && errno == EINTR);
  }
  Glib::usleep(50000);
}

std::pair<const char*, char* const*> FaxListDialog::get_ps_viewer_parms(const char* filename) {

  // lock the Prog_config object to stop it being modified in the intial (GUI) thread
  // while we are accessing it here
  Glib::Mutex::Lock lock(*prog_config.mutex_p);

  std::vector<std::string> view_parms;
  std::string view_cmd;
  std::string view_name;
  std::string::size_type end_pos;
  try {
    view_cmd = Glib::filename_from_utf8(prog_config.ps_view_cmd);
  }
  catch (Glib::ConvertError&) {
    write_error("UTF-8 conversion error in FaxListDialog::ps_viewer()\n");
    return std::pair<const char*, char* const*>(0,0);
  }

  if ((end_pos = view_cmd.find_first_of(' ')) != std::string::npos) { // we have parms
    view_name.assign(view_cmd, 0, end_pos);
    view_parms.push_back(view_name);
    // find start of next parm
    std::string::size_type start_pos = view_cmd.find_first_not_of(' ', end_pos);
    while (start_pos != std::string::npos) {
      end_pos = view_cmd.find_first_of(' ', start_pos);
      if (end_pos != std::string::npos) {
	view_parms.push_back(view_cmd.substr(start_pos, end_pos - start_pos));
	start_pos = view_cmd.find_first_not_of(' ', end_pos); // prepare for next interation
      }
      else {
	view_parms.push_back(view_cmd.substr(start_pos, 
					     view_cmd.size() - start_pos));
	start_pos = end_pos;
      }
    }
  }

  else { // just a print command without parameters to be passed
    view_name = view_cmd;
    view_parms.push_back(view_name);
  }

  view_parms.push_back(filename);

  char** exec_parms = new char*[view_parms.size() + 1];

  char** temp_pp = exec_parms;
  std::vector<std::string>::const_iterator iter;
  for (iter = view_parms.begin(); iter != view_parms.end(); ++iter, ++temp_pp) {
    *temp_pp = new char[iter->size() + 1];
    std::strcpy(*temp_pp, iter->c_str());
  }
  
  *temp_pp = 0;

  char* prog_name = new char[view_name.size() + 1];
  std::strcpy(prog_name, view_name.c_str());

  return std::pair<const char*, char* const*>(prog_name, exec_parms);
}

void FaxListDialog::print_fax(void) {

  Gtk::TreeModel::iterator row_iter = tree_view.get_selection()->get_selected();
  if (row_iter) {
    // fax_basename_sem is initialised to a count of 1, so this will stop
    // fax_basename being overwritten until Sem_sync::post() is called in
    // print_fax_thread() or view_fax_thread() or on Glib::ThreadError being
    // caught
    fax_basename_sem.wait();

    fax_basename = prog_config.homedir;
    if (mode == received) fax_basename += "/faxin/";
    else fax_basename += "/faxsent/";

    fax_basename += (*row_iter)[model_columns.fax_name];
    fax_basename += '/';
    fax_basename += (*row_iter)[model_columns.fax_name];
    fax_basename += '.';

    // now block off the signals for which we have set handlers so that the socket server
    // thread does not receive the signals, otherwise we will have memory synchronisation
    // issues in multi-processor systems - we will unblock in the initial (GUI) thread
    // as soon as the socket server thread has been launched
    sigset_t sig_mask;
    sigemptyset(&sig_mask);
    sigaddset(&sig_mask, SIGCHLD);
    sigaddset(&sig_mask, SIGQUIT);
    sigaddset(&sig_mask, SIGTERM);
    sigaddset(&sig_mask, SIGINT);
    sigaddset(&sig_mask, SIGHUP);
#ifdef HAVE_PTHREAD_SIGMASK 
    pthread_sigmask(SIG_BLOCK, &sig_mask, 0);
#else
    sigprocmask(SIG_BLOCK, &sig_mask, 0);
#endif

    try {
      Glib::Thread::create(sigc::mem_fun(*this, &FaxListDialog::print_fax_thread), false);
    }
    catch (Glib::ThreadError&) {
      write_error("Cannot start new print thread, fax will not run be printed\n");
      fax_basename_sem.post();
    }
    // now unblock the signals so that the initial (GUI) thread can receive them
#ifdef HAVE_PTHREAD_SIGMASK 
    pthread_sigmask(SIG_UNBLOCK, &sig_mask, 0);
#else
    sigprocmask(SIG_UNBLOCK, &sig_mask, 0);
#endif
  }
}

void FaxListDialog::print_fax_thread(void) {

  // get the arguments for the exec() calls below (because this is a
  // multi- threaded program, we must do this before fork()ing because
  // we use functions to get the arguments which are not async-signal-safe)
  std::pair<const char*, char* const*> print_from_stdin_parms(get_print_from_stdin_parms());
  std::pair<const char*, char* const*> fax_to_ps_parms(get_fax_to_ps_parms(fax_basename, true));

  // release main GUI thread in print_fax() or view_fax()
  // waiting for this use of fax_basename to finish
  fax_basename_sem.post();

  if (print_from_stdin_parms.first) { // this will be 0 if get_print_from_stdin_parms()
                                      // threw a Glib::ConvertError)

    // now launch a new process to control the printing process
    // the main program process needs to continue while the printing
    // is going on
    pid_t pid = fork();

    if (pid == -1) {
      write_error("Fork error - exiting\n");
      std::exit(FORK_ERROR);
    }
    if (!pid) { // child print process

      // make the call to connect_to_stderr() in this child process
      // dependent on calling a print confirmation popupr dialog: if we have
      // elected not to bring up such a dialog, we are probably using a
      // print manager such as kprinter, and we can get confusing messages
      // from stderr from these
      // now mutex is required for the access to prog_config.print_popup
      // because we have fork()ed into a separate process
      if (prog_config.print_popup) connect_to_stderr();

      // now create a blocking pipe which the print processes can use to
      // communicate via stdout/stdin
      Pipe_fifo fork_pipe(Pipe_fifo::block);
      
      pid_t pid = fork();

      if (pid == -1) {
	write_error("Fork error - exiting\n");
	_exit(FORK_ERROR);
      }
      if (!pid) { // process to exec to print_from_stdin()
      
	fork_pipe.connect_to_stdin();
	execvp(print_from_stdin_parms.first, print_from_stdin_parms.second);
	// if we reached this point, then the execvp() call must have failed
	// write error and then end process - use _exit(), not exit()
	write_error("Can't find the print program - please check your installation\n"
		    "and the PATH environmental variable\n");
	_exit(0);
      }

      // this is the process which will send postscript to stdout
      fork_pipe.connect_to_stdout();
      execvp(fax_to_ps_parms.first, fax_to_ps_parms.second);

      // if we reached this point, then the execvp() call must have failed
      // report error and then end process - use _exit(), not exit()
      write_error("Can't find the efix-0.9a program - please check your installation\n"
		  "and the PATH environmental variable\n");
      _exit(0);
    }
    // release the memory allocated on the heap for
    // the redundant print_from_stdin_parms and fax_to_ps_parms
    // we are in the main parent process here - no worries about
    // only being able to use async-signal-safe functions
    delete_parms(print_from_stdin_parms);
  }
  delete_parms(fax_to_ps_parms);
}


void FaxListDialog::view_fax(void) {
 
  Gtk::TreeModel::iterator row_iter = tree_view.get_selection()->get_selected();
  if (row_iter && !prog_config.ps_view_cmd.empty()) {
    // fax_basename_sem is initialised to a count of 1, so this will stop
    // fax_basename being overwritten until Sem_sync::post() is called in
    // print_fax_thread() or view_fax_thread() or on Glib::ThreadError being
    // caught
    fax_basename_sem.wait();

    fax_basename = prog_config.homedir;
    if (mode == received) fax_basename += "/faxin/";
    else fax_basename += "/faxsent/";

    fax_basename += (*row_iter)[model_columns.fax_name];
    fax_basename += '/';
    fax_basename += (*row_iter)[model_columns.fax_name];
    fax_basename += '.';

    // now block off the signals for which we have set handlers so that the socket server
    // thread does not receive the signals, otherwise we will have memory synchronisation
    // issues in multi-processor systems - we will unblock in the initial (GUI) thread
    // as soon as the socket server thread has been launched
    sigset_t sig_mask;
    sigemptyset(&sig_mask);
    sigaddset(&sig_mask, SIGCHLD);
    sigaddset(&sig_mask, SIGQUIT);
    sigaddset(&sig_mask, SIGTERM);
    sigaddset(&sig_mask, SIGINT);
    sigaddset(&sig_mask, SIGHUP);
#ifdef HAVE_PTHREAD_SIGMASK 
    pthread_sigmask(SIG_BLOCK, &sig_mask, 0);
#else
    sigprocmask(SIG_BLOCK, &sig_mask, 0);
#endif

    try {
      Glib::Thread::create(sigc::mem_fun(*this, &FaxListDialog::view_fax_thread), false);
    }
    catch (Glib::ThreadError&) {
      write_error("Cannot start new view fax thread, fax cannot be viewed\n");
      fax_basename_sem.post();
    }
    // now unblock the signals so that the initial (GUI) thread can receive them
#ifdef HAVE_PTHREAD_SIGMASK 
    pthread_sigmask(SIG_UNBLOCK, &sig_mask, 0);
#else
    sigprocmask(SIG_UNBLOCK, &sig_mask, 0);
#endif
  }
}

void FaxListDialog::view_fax_thread(void) {
 
  // get a temporary file
  char filename[] = "/tmp/efax-gtk-view.XXXXXX";
  int file_fd = mkstemp(filename);

  if (file_fd == -1) {
    write_error("Failed to make temporary file:\n"
		"please check permissions in /tmp directory");
    _exit(0);
  }

  // now create a pipe and proceed to fork to write the postscript to temporary file
  Pipe_fifo fork_pipe(Pipe_fifo::block);

  // get the arguments for the exec() call to convert the fax files to
  // PS format following the next fork() call (because this is a multi-
  // threaded program, we must do this before fork()ing because we use
  // functions to get the arguments which are not async-signal-safe)
  std::pair<const char*, char* const*> fax_to_ps_parms(get_fax_to_ps_parms(fax_basename, false));

  // release main GUI thread in print_fax() or view_fax()
  // waiting for this use of fax_basename to finish
  fax_basename_sem.post();

  // now fork to create the process which will write postscript to stdout
  pid_t pid = fork();

  if (pid == -1) {
    write_error("Fork error - exiting\n");
    _exit(FORK_ERROR);
  }
    
  if (!pid) { // child process which will send postscript to stdout
  
    // this is the child process to write to stdin
    // now we have fork()ed we can connect to stderr
    connect_to_stderr();

    fork_pipe.connect_to_stdout();
    execvp(fax_to_ps_parms.first, fax_to_ps_parms.second);

    // if we reached this point, then the execvp() call must have failed
    // report error and then end process - use _exit(), not exit()
    write_error("Can't find the efix-0.9a program - please check your installation\n"
		"and the PATH environmental variable\n");
    _exit(0);
  }
  // this is the parent process, which will convert the fax to Postscript
  // and then pipe it on stdout to write_from_stdin()
    
  // first, release the memory allocated on the heap for
  // the redundant fax_to_ps_parms
  // we are in the main parent process here - no worries about
  // only being able to use async-signal-safe functions
  delete_parms(fax_to_ps_parms);
    
  // now write stdin in (containing the generated PS text) to file
  fork_pipe.connect_to_stdin();
  write_from_stdin(file_fd);

  // the temporary file to be viewed has now been created
  // so launch a new process to control the viewing process
  // we will reset handlers and then fork() again having done
  // so, so that we can wait on the child of child

  // first get the arguments for the exec() call to launch the fax
  // viewer following the next fork() call (because this is a multi-
  // threaded program, we must do this before fork()ing because we use
  // functions to get the arguments which are not async-signal-safe)
  std::pair<const char*, char* const*> ps_viewer_parms(get_ps_viewer_parms(filename));

  if (ps_viewer_parms.first) { // this will be 0 if get_ps_viewer_parms()
                               // threw a Glib::ConvertError)
    
    pid = fork();
    
    if (pid == -1) {
      write_error("Fork error\n");
      std::exit(FORK_ERROR);
    }
    if (!pid) { // child view-control process
    
      // restore the default SIGCHLD signal handler with SA_RESTART
      // for systems (such as Linux) which require it.  Also unblock
      // signals in case we are blocking them on launching a thread
      // elsewhere in this program (I cannot work out from IEEE 1003.1
      // whether all masks are emptied on a fork() in any event)
      sigset_t sig_mask;
      sigemptyset(&sig_mask);
      sigaddset(&sig_mask, SIGCHLD);
      sigaddset(&sig_mask, SIGQUIT);
      sigaddset(&sig_mask, SIGTERM);
      sigaddset(&sig_mask, SIGINT);
      sigaddset(&sig_mask, SIGHUP);
      // this child process is single threaded, so we can use sigprocmask()
      // rather than pthread_sigmask() (and should do so as sigprocmask()
      // is guaranteed to be async-signal-safe)
      sigprocmask(SIG_UNBLOCK, &sig_mask, 0);

      struct sigaction sig_act;
      sig_act.sa_handler = SIG_DFL;
      sigemptyset(&sig_act.sa_mask);
#ifdef SA_RESTART
      sig_act.sa_flags = SA_RESTART;
#else
      sig_act.sa_flags = 0;
#endif
      sigaction(SIGCHLD, &sig_act, 0);
      
      connect_to_stderr();

      // now fork() again
      pid_t pid = fork();

      if (pid == -1) {
	write_error("Fork error - exiting\n");
	_exit(FORK_ERROR);
      }
    
      if (!pid) { // child process to call ps_viewer() from
	  
	execvp(ps_viewer_parms.first, ps_viewer_parms.second);

	// if we reached this point, then the execvp() call must have failed
	// report error and then end process - use _exit(), not exit()
	write_error("Can't find the ps viewer program - please check your installation\n"
		    "and the PATH environmental variable\n");
	_exit(0);
      }
      // now we are back to the main viewing process again
      // delete the temporary file

      // first wait till the ps_viewer() process is closed by the user exiting
      // the postscript viewing program
      wait(0);
    
      unlink(filename);
      // now terminate the process
      _exit(0);
    }
    // release the memory allocated on the heap for
    // the redundant ps_viewer_parms
    // we are in the main parent process here - no worries about
    // only being able to use async-signal-safe functions
    delete_parms(ps_viewer_parms);
  }
}

void FaxListDialog::delete_parms(std::pair<const char*, char* const*> parms_pair) {

  delete[] parms_pair.first;
  char* const* temp_pp = parms_pair.second;
  for(; *temp_pp != 0; ++temp_pp) {
    delete[] *temp_pp;
  }
  delete[] parms_pair.second;
}

void FaxListDialog::delete_fax_prompt(void) {

  Gtk::TreeModel::iterator row_iter = tree_view.get_selection()->get_selected();
  if (row_iter) {
    std::string faxnumber((*row_iter)[model_columns.fax_name]);
    Glib::ustring msg(gettext("Delete fax "));
    msg += faxnumber + (gettext("?"));
    
    PromptDialog* dialog_p = new PromptDialog(msg, gettext("Delete fax"), standard_size, *this);
    dialog_p->accepted.connect(sigc::mem_fun(*this, &FaxListDialog::delete_fax));
    // there is no memory leak -- the memory will be deleted when PromptDialog closes
  }
}

void FaxListDialog::delete_fax(void) {

  Gtk::TreeModel::iterator row_iter = tree_view.get_selection()->get_selected();
  if (row_iter) {
    // get the name of the fax to be moved and the new name
    std::string old_dirname(prog_config.homedir);
    std::string new_dirname(prog_config.homedir);
    if (mode == received) {
      old_dirname += "/faxin/";
      new_dirname += "/faxin/oldfax/";
    }
    else {
      old_dirname += "/faxsent/";
      new_dirname += "/faxsent/oldfax/";
    }
    old_dirname += (*row_iter)[model_columns.fax_name];
    new_dirname += (*row_iter)[model_columns.fax_name];

    // make a new directory to copy the files into
    mkdir(new_dirname.c_str(), S_IRUSR | S_IWUSR | S_IXUSR);

    // it should be possible to use link()/unlink() as the target and source
    // are sub-directories of $HOME/faxin or $HOME/faxsent.  If because of a
    // very odd filesystem arrangement this won't work, the method will do nothing

    std::vector<std::string> filelist;
    struct dirent* direntry;
    struct stat statinfo;

    DIR* dir_p;
    if ((dir_p = opendir(old_dirname.c_str())) == 0) {
      std::string msg("Can't open directory ");
      msg += old_dirname;
      msg += '\n';
      write_error(msg.c_str());
    }

    else {
      chdir(old_dirname.c_str());
      while ((direntry = readdir(dir_p)) != 0) {
	stat(direntry->d_name, &statinfo);
	if (S_ISREG(statinfo.st_mode)) {
	  filelist.push_back(direntry->d_name);
	}
      }

      closedir(dir_p);

      bool failure = false;
      std::vector<std::string>::const_iterator iter;
      std::string old_filename;
      std::string new_filename;

      for (iter = filelist.begin(); iter != filelist.end(); ++iter) {
	old_filename = old_dirname + '/';
	old_filename += *iter;
	new_filename = new_dirname + '/';
	new_filename += *iter;
	if (link(old_filename.c_str(), new_filename.c_str())) {
	  failure = true;
	  break;
	}
      }

      if (failure) { // recover position
	std::vector<std::string>::const_iterator recover_iter;
	for (recover_iter = filelist.begin(); recover_iter != iter; ++recover_iter) {
	  new_filename = new_dirname + '/';
	  new_filename += *recover_iter;
	  unlink(new_filename.c_str());
	}
	rmdir(new_dirname.c_str());
	std::string msg("Can't move directory ");
	msg += old_dirname;
	msg += " to ";
	msg += new_dirname;
	msg += '\n';
	write_error(msg.c_str());
      }
      else {
	for (iter = filelist.begin(); iter != filelist.end(); ++iter) {
	  old_filename = old_dirname + '/';
	  old_filename += *iter;
	  unlink(old_filename.c_str());
	}
	if (rmdir(old_dirname.c_str())) {
	  std::string msg("Can't delete directory ");
	  msg += old_dirname;
	  msg += "\nThe contents should have been moved to ";
	  msg += new_dirname;
	  msg += "\nand it should now be empty -- please check\n";
	  write_error(msg.c_str());
	}
	else get_fax_list_rows();
      }
      // reset current directory
      std::string temp(prog_config.homedir + "/faxin");
      chdir(temp.c_str());
    }
  }
}

void FaxListDialog::refresh_slot(void) {

  get_fax_list_rows();
  set_buttons_slot();
}

DescriptionDialog::DescriptionDialog(const int standard_size, const std::string& text, Gtk::Window& window):
                             Gtk::Window(Gtk::WINDOW_TOPLEVEL), in_run_loop(false),
			     ok_button(Gtk::Stock::OK), cancel_button(Gtk::Stock::CANCEL),
			     button_box(Gtk::BUTTONBOX_END, standard_size/2),
			     label(gettext("Fax description?")), table(3, 1, false),
                             parent(window) {

  button_box.add(cancel_button);
  button_box.add(ok_button);

  table.attach(label, 0, 1, 0, 1, Gtk::FILL | Gtk::EXPAND,
	 Gtk::FILL | Gtk::EXPAND, standard_size/2, standard_size/4);

  table.attach(entry, 0, 1, 1, 2, Gtk::FILL | Gtk::EXPAND,
	 Gtk::EXPAND, standard_size/2, standard_size/4);

  table.attach(button_box, 0, 1, 2, 3, Gtk::FILL | Gtk::EXPAND,
	 Gtk::SHRINK, standard_size/2, standard_size/4);

  ok_button.signal_clicked().connect(sigc::bind(sigc::mem_fun(*this, &DescriptionDialog::selected), true));
  cancel_button.signal_clicked().connect(sigc::bind(sigc::mem_fun(*this, &DescriptionDialog::selected), false));

  ok_button.set_flags(Gtk::CAN_DEFAULT);
  cancel_button.set_flags(Gtk::CAN_DEFAULT);

  add(table);
  
  entry.set_text(text);
  set_title(gettext("Fax description"));
  set_transient_for(parent);
  set_type_hint(Gdk::WINDOW_TYPE_HINT_DIALOG);
  parent.set_sensitive(false);
  set_modal(true);

  entry.set_size_request(standard_size * 9, standard_size);
  set_border_width(standard_size/2);

  entry.grab_focus();

  set_position(Gtk::WIN_POS_CENTER_ON_PARENT);
  set_resizable(false);

  set_icon(Gdk::Pixbuf::create_from_xpm_data(window_icon_xpm));

  show_all();

  // we now need to deselect what is in the entry
  entry.select_region(0,0);
}

void DescriptionDialog::run(void) {
  in_run_loop = true;
  Gtk::Main::run();
}

void DescriptionDialog::selected(bool accept) {
  parent.set_sensitive(true); // do this before we emit accepted()
  hide_all();
  if (accept) accepted(entry.get_text());
  if (in_run_loop) Gtk::Main::quit();
  // if we have not called run(), then this dialog is self-owning and it is safe to call `delete this'
  else delete this;
}

bool DescriptionDialog::on_delete_event(GdkEventAny*) {
  selected(false);
  return true; // returning true prevents destroy sig being emitted
}

bool DescriptionDialog::on_key_press_event(GdkEventKey* event_p) {

  if (event_p->keyval == GDK_Escape) selected(false);
  else if (event_p->keyval == GDK_Return && !cancel_button.has_focus()) selected(true);
  else Gtk::Window::on_key_press_event(event_p);
  return true; // processing ends here
}
